import io
import base64
import requests
import pandas as pd
import os
from .write import ViWriteClient
from .analytics import ViAnalyticsMixin
from .utils import decorate_functions_by_argument, set_default_collection
from .errors import LoginError, APIError

class ViClient(ViWriteClient, ViAnalyticsMixin):
    """
        The main Vi client with most of the available read and write methods available to it.

        Parameters:
            username:
                your username for accessing vectorai
            api_key:
                your api key for accessing vectorai
            url:
                url of the deployed vectorai database

        Example:
            >>> from vectorai.client import ViClient
            >>> vi_client = ViClient(username, api_key, vectorai_url)
            >>> vi_client.list_collections()
    """

    def __init__(self, username: str=None, api_key: str=None, url: str = "https://api.vctr.ai", verbose: bool = True) -> None:
        super().__init__(username, api_key, url)
        if username is None:
            if 'VI_USERNAME' not in os.environ.keys():
                raise APIError("Specify username of set VI_USERNAME as an environment variable.")
            username = os.environ['VI_USERNAME']
        
        if api_key is None:
            if 'VI_API_KEY' not in os.environ.keys():
                raise APIError("Specify VectorAI API key VI_API_KEY as an environment variable.")
            api_key = os.environ['VI_API_KEY']
        
        self.username = username
        self.api_key = api_key
        self.url = url

        if verbose:
            self.check_login_details()
            print(
                f"Logged in. Welcome {self.username}. To view list of available collections, call list_collections() method."
            )
        
    def check_login_details(self):
        try:
            self.list_collections()
        except:
            raise LoginError("Username, api key or url is incorrect.")


def request_api_key(username: str, email:str, description:str="I'd like to try it out.", referral_code: str="github_referred"):
    """
        Request an api key
        Make sure to save the api key somewhere safe. If you have a valid referral code, you can recieve the api key more quickly.
            
        Args:
            username:
                Username you'd like to create, lowercase only
            email:
                Email you are using to sign up
            description:
                Description of your intended use case
            referral_code:
                The referral code you've been given to allow you to register for an api key before others
    """
    return requests.post(
        url="{}/project/request_api_key".format("https://api.vctr.ai"),
        json={
            "username": username,
            "email": email,
            "referral_code": referral_code,
            "description": description,
        },
    ).json()

@decorate_functions_by_argument(set_default_collection, 'collection_name')
class ViCollectionClient(ViClient):
    """
        The Vi client when you are mainly working with 1 client.

        Args:
            username:
                your username for accessing vecdb
            api_key:
                your api key for accessing vecdb
            url:
                url of the deployed vecdb database
            collection_name:
                The name of the collection

        Example:
            >>> from vectorai.client import ViClient
            >>> vi_client = ViClient(username, api_key, collection_name, vectorai_url)
            >>> vi_client.insert_documents(documents)
    """
    def __init__(self, collection_name: str, username: str, api_key: str, url: str="https://api.vctr.ai", verbose: bool=True) -> None:
        if username is None:
            if 'VI_USERNAME' not in os.environ.keys():
                raise APIError("Specify username of set VI_USERNAME as an environment variable.")
            username = os.environ['VI_USERNAME']
        
        if api_key is None:
            if 'VI_API_KEY' not in os.environ.keys():
                raise APIError("Specify VectorAI API key VI_API_KEY as an environment variable.")
            api_key = os.environ['VI_API_KEY']
        
        self.username = username
        self.api_key = api_key
        self._collection_name = collection_name
        self.url = url
        if verbose:
            self.check_login_details()
            print(
                f"Logged in. Welcome {self.username}. To view list of available collections, call list_collections() method."
            )
            setattr(self, 'decorator_called', False)

    @property
    def collection_name(self) -> str:
        return self._collection_name
    
    @collection_name.setter
    def collection_name(self, value: str) -> None:
        self._collection_name = value
        
    @collection_name.getter
    def collection_name(self) -> str:
        return self._collection_name



"""Document-Specific Utilities
"""
from typing import List, Dict, Any
from .errors import MissingFieldError

class DocUtilsMixin:
    @classmethod
    def get_field(self, field: str, doc: Dict):
        """
            For nested dictionaries, tries to access a field.
            e.g. 
            field = kfc.item
            This should return "chickens" based on doc below.
            {
                "kfc": {
                    "item": "chickens"
                }
            }

            Args:
                field:
                    Field of a document.
                doc: 
                    document

            Example:
                >>> from vectorai.client import ViClient
                >>> vi_client = ViClient(username, api_key, vectorai_url)
                >>> sample_document = {'kfc': {'item': 'chicken'}}
                >>> vi_client.get_field('kfc.item', sample_document) == 'chickens'
        """
        d = doc
        for f in field.split("."):
            try:
                d = d[f]
            except:
                try:
                    return doc[field]
                except:
                    raise MissingFieldError("Document is missing " + field)
        return d

    @classmethod
    def get_fields(self, fields: List[str], doc: Dict) -> List[Any]:
        """
            For nested dictionaries, tries to access a field.
            e.g. 
            field = kfc.item
            This should return "chickens" based on doc below.
            {
                "kfc": {
                    "item": "chickens"
                }
            }

            Args:
                fields:
                    List of fields of a document.
                doc: 
                    document

            Example:
                >>> from vectorai.client import ViClient
                >>> vi_client = ViClient(username, api_key, vectorai_url)
                >>> sample_document = {'kfc': {'item': 'chicken'}}
                >>> vi_client.get_field('kfc.item', sample_document) == 'chickens'
        """
        return [self.get_field(f, doc) for f in fields]

    def get_field_across_documents(self, field: str, docs: List[Dict]):
        """
            For nested dictionaries, tries to access a field.
            e.g. 
            field = kfc.item
            This should return "chickens" based on doc below.
            {
                "kfc": {
                    "item": "chickens"
                }
            }

            Args:
                fields:
                    List of fields of a document.
                doc: 
                    document

            Example:
                >>> from vectorai.client import ViClient
                >>> vi_client = ViClient(username, api_key, vectorai_url)
                >>> documents = vi_client.create_sample_documents(10)
                >>> vi_client.get_field_across_documents('size.cm', documents)
                # returns 10 values in the nested dictionary
        """
        return [self.get_field(field, doc) for doc in docs]

    @staticmethod
    def set_field(
        field: str, doc: Dict, value: Any, handle_if_missing=True
    ):
        """
        For nested dictionaries, tries to write to the respective field.
        If you toggle off handle_if_misisng, then it will output errors if the field is 
        not found.
        e.g.
        field = kfc.item
        value = "fried wings"
        This should then create the following entries if they dont exist:
        {
            "kfc": {
                "item": "fried wings"
            }
        }

        Args:
            field:
                Field of the document to write.
            doc: 
                Python dictionary
            value:
                Value to write

        Example:

            >>> from vectorai.client import ViClient
            >>> vi_client = ViClient(username, api_key, vectorai_url)
            >>> sample_document = {'kfc': {'item': ''}}
            >>> vi_client.set_field('kfc.item', sample_document, 'chickens')
        """
        fields = field.split(".")
        # Assign a pointer.
        d = doc
        for i, f in enumerate(fields):
            # Assign the value if this is the last entry e.g. stores.fastfood.kfc.item will be item 
            if i == len(fields) - 1:
                d[f] = value
            else:
                if f in d.keys():
                    d = d[f]
                else:
                    d.update({f: {}})
    
    def set_field_across_documents(
        self, field: str, values: List[Any], docs: List[Dict]
    ):
        """
        For multiple documents, set the right fields.
        e.g.
        field = kfc.item
        value = "fried wings"
        This should then create the following entries if they dont exist:
        {
            "kfc": {
                "item": "fried wings"
            }
        }

        Args:
            field:
                Field of the document to write.
            doc: 
                Python dictionary
            value:
                Value to write

        Example:

            >>> from vectorai.client import ViClient
            >>> vi_client = ViClient(username, api_key, vectorai_url)
            >>> sample_document = {'kfc': {'item': ''}}
            >>> vi_client.set_fields('kfc.item', sample_document, 'chickens')
        """
        assert len(values) == len(docs), "Assert that the number of values " + \
            "equates to the number of documents"
        for i, value in enumerate(values):
            self.set_field(field, docs[i], value)

    @staticmethod
    def is_field(field: str, doc: Dict) -> bool:
        """
        For nested dictionaries, tries to access a field.
        e.g. 
        field = kfc.item
        This should return "chickens" based on doc below.
        {
            "kfc": {
                "item": "chickens"
            }
        }

        Args:
            collection_name:
                Name of collection.
            job_id: 
                ID of the job.
            job_name:
                Name of the job.

        Example:
            >>> from vectorai.client import ViClient
            >>> vi_client = ViClient(username, api_key, vectorai_url)
            >>> sample_document = {'kfc': {'item': 'chicken'}}
            >>> vi_client.is_field('kfc.item', sample_document) == True
        """
        d = doc
        for f in field.split("."):
            try:
                d = d[f]
            except:
                return False
        return True



"""Reads Operations designed for python
"""
import io
import base64
import requests
import random
import pandas as pd
import time
import warnings
from typing import List, Dict, Union, Any
from .api.read import ViReadAPIClient
from .utils import UtilsMixin
from .doc_utils import DocUtilsMixin
from .errors import MissingFieldWarning
class ViReadClient(ViReadAPIClient, UtilsMixin, DocUtilsMixin):
    def __init__(self, username: str, api_key: str, url: str="https://api.vctr.ai"):
        self.username = username
        self.api_key = api_key
        self.url = url

    def random_aggregation_query(
        self, collection_name: str, groupby: int = 1, metrics: int = 1
    ):
        """
        Generates a random filter query.

        Args:
            collection_name:
                name of collection
            groupby:
                The number of groupbys to randomly generate
            metrics:
                The number of metrics to randomly generate

        Example:
            >>> from vectorai.client import ViClient
            >>> vi_client = ViClient(username, api_key, vectorai_url)
            >>> vi_client.random_aggregation_query(collection_name, groupby=1, metrics=1)
        """
        schema = self.collection_schema(collection_name)
        full_aggregation_query = {"groupby": [], "metrics": []}
        for s in schema:
            if schema[s] == "text":
                full_aggregation_query["groupby"].append(
                    {"name": s, "field": s, "agg": "texts"}
                )
            elif schema[s] == "numeric":
                full_aggregation_query["metrics"].append(
                    {"name": s, "field": s, "agg": "avg"}
                )
        return {
            "groupby": random.sample(full_aggregation_query["groupby"], groupby),
            "metrics": random.sample(full_aggregation_query["metrics"], metrics),
        }

    def search(self,
        collection_name: str, 
        vector: List,
        field: List,
        approx: int = 0,
        sum_fields: bool = True,
        metric: str = "cosine",
        min_score=None,
        page: int = 1,
        page_size: int = 10,
        include_vector:bool=False,
        include_count:bool=True,
        asc:bool=False,
    ):
        """
Vector Similarity Search. Search a vector field with a vector, a.k.a Nearest Neighbors Search

Enables machine learning search with vector search. Search with a vector for the most similar vectors.

For example: Search with a person's characteristics, who are the most similar (querying the "persons_characteristics_vector" field)::

    Query person's characteristics as a vector: 
    [180, 40, 70] representing [height, age, weight]

    Search Results:
    [
        {"name": Adam Levine, "persons_characteristics_vector" : [180, 56, 71]},
        {"name": Brad Pitt, "persons_characteristics_vector" : [180, 56, 65]},
    ...]

Args:
	vector:
		Vector, a list/array of floats that represents a piece of data.
	collection_name:
		Name of Collection
	search_fields:
		Vector fields to search through
	approx:
		Used for approximate search
	sum_fields:
		Whether to sum the multiple vectors similarity search score as 1 or seperate
	page_size:
		Size of each page of results
	page:
		Page of the results
	metric:
		Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
	min_score:
		Minimum score for similarity metric
	include_vector:
		Include vectors in the search results
	include_count:
		Include count in the search results
    asc:
        Whether to sort the score by ascending order (default is false, for getting most similar results)
"""                
                
        search_fields ={}
        if isinstance(field, str):
            advanced_search_query = {
                field.replace('_vector_', ''): {'vector': vector, 'fields': [field]}
            }
        else:
            advanced_search_query = {
                field[0].replace('_vector_', ''): {'vector': vector, 'fields': field}
            }
        return self.advanced_search(
            collection_name=collection_name,
            multivector_query=advanced_search_query,
            approx=approx,
            sum_fields=sum_fields,
            metric=metric,
            min_score=min_score,
            page=page,
            page_size=page_size,
            include_vector=include_vector,
            include_count=include_count,
            asc=asc
        )

    def random_filter_query(
        self, collection_name: str, text_filters: int = 1, numeric_filters: int = 0
    ):
        """
        Generates a random filter query.

        Args:
            collection_name:
                name of collection
            text_filters:
                The number of text filters to randomly generate
            numeric_filters:
                The number of numeric filters to randomly generate

        Example:
            >>> from vectorai.client import ViClient
            >>> vi_client = ViClient(username, api_key, vectorai_url)
            >>> vi_client.random_filter_query(collection_name, text_filters=1, numeric_filters=0)
        """
        schema = self.collection_schema(collection_name)
        facets = self.facets(collection_name)
        full_filter_query = {"text": [], "numeric": []}
        for f, t in schema.items():
            if t == "text":
                if isinstance(facets[f], list):
                    full_filter_query["text"].append(
                        {
                            "field": f,
                            "filter_type": "text",
                            "condition_value": random.sample(facets[f], 1)[0][f],
                            "condition": "==",
                        }
                    )
            elif t == "numeric":
                if isinstance(facets[f], dict):
                    full_filter_query["numeric"].append(
                        {
                            "field": f,
                            "filter_type": "date",
                            "condition_value": (facets[f]["max"] - facets[f]["min"])
                            / 2,
                            "condition": ">=",
                        }
                    )
        return random.sample(full_filter_query["text"], text_filters) + random.sample(
            full_filter_query["numeric"], numeric_filters
        )

    def head(
        self, collection_name: str, page_size: int = 5, return_as_pandas_df: bool = True
    ):
        """
        The main Vi client with most of the available read and write methods available to it.

        Args:
            collection_name:
                The name of your collection
            page_size:
                The number of results to return
            return_as_pandas_df:
                If True, return as a pandas DataFrame rather than a JSON.

        Example:
            >>> from vectorai.client import ViClient
            >>> vi_client = ViClient(username, api_key, vectorai_url)
            >>> vi_client.head(collection_name, page_size=10)
        """
        response = requests.get(
            url="{}/collection/retrieve_documents".format(self.url),
            params={
                "username": self.username,
                "api_key": self.api_key,
                "collection_name": collection_name,
                "page_size": page_size,
            },
        ).json()
        if "documents" in response.keys():
            response = response["documents"]
        if return_as_pandas_df:
            return pd.DataFrame.from_records(response)
        else:
            return response

    def retrieve_all_documents(
        self,
        collection_name: str,
        sort_by: List = [],
        asc: bool = True,
        include_vector: bool = True,
        include_fields: List = [],
        retrieve_chunk_size: int=1000
    ):
        """
        Retrieve all documents in a given collection. We recommend specifying specific fields to extract
        as otherwise this function may take a long time to run.

        Args:
            collection_name:
                Name of collection.
            sort_by:
                Select the fields by which to sort by.
            asc:
                If true, returns in ascending order of what is sort.
            include_vector:
                If true, includes _vector_ fields to return them.
            include_fields:
                Adjust which fields are returned.
            retrieve_chunk_size:
                The number of documents to retrieve per request.

        Example:
            >>> from vectorai.client import ViClient
            >>> vi_client = ViClient(username, api_key, vectorai_url)
            >>> all_documents = vi_client.retrieve_all_documents(collection_name)
        """
        num_of_docs = self.collection_stats(collection_name)['number_of_documents']
        with self.progress_bar(list(range(int(num_of_docs/ retrieve_chunk_size)))) as pbar:
            d = self.retrieve_documents(
                collection_name, retrieve_chunk_size, sort=sort_by, asc=asc, include_vector=include_vector, 
                include_fields=include_fields
            )
            all_docs = d["documents"]
            pbar.update(1)
            while len(d["documents"]) > 0:
                d = self.retrieve_documents(
                    collection_name,
                    retrieve_chunk_size,
                    d["cursor"],
                    sort=sort_by,
                    asc=asc,
                    include_vector=include_vector,
                    include_fields=include_fields
                )
                all_docs += d["documents"]
                pbar.update(1)
        return all_docs

    def wait_till_jobs_complete(self, collection_name: str, job_id: str, job_name: str):
        """
        Wait until a specific job is complete.

        Args:
            collection_name:
                Name of collection.
            job_id: 
                ID of the job.
            job_name:
                Name of the job.

        Example:
            >>> from vectorai.client import ViClient
            >>> vi_client = ViClient(username, api_key, vectorai_url)
            >>> job = vi_client.dimensionality_reduction_job('nba_season_per_36_stats_demo', vector_field='season_vector_', n_components=2)
            >>> vi_client.wait_till_jobs_complete('nba_season_per_36_stats_demo', **job)

        """
        status = self.job_status(collection_name, job_id, job_name)
        while (
            status["status"] == "Running"
            or status["status"] == "Started"
            or status["status"] == "NotStarted"
        ):
            status = self.job_status(collection_name, job_id, job_name)
            time.sleep(15)
        print(status)
        return "Done"

    def check_schema(self, collection_name: str, document: Dict=None):
        """
        Check the schema of a given collection.

        Args:
            collection_name:
                Name of collection.

        Example:
            >>> from vectorai.client import ViClient
            >>> vi_client = ViClient(username, api_key, vectorai_url)
            >>> vi_client.check_schema(collection_name)
        """
        if document is None:
            document = self.retrieve_documents(collection_name, page_size=1)
        self._check_schema(document)

    def _check_schema(
        self,
        document: Dict,
        is_missing_vector_field=True,
        is_missing_id_field=True,
        is_nested=False
    ):
        """
        Check if there is a _vector_ field and an _id field.

        Args:
            document:
                A JSON file/python dictionary
            is_missing_vector_field:
                DO NOT CHANGE. A tracker to return if the dictionary is missing a vector field
            is_nested:
                DO NOT CHANGE. Returns True if is a nested. Used internally for recursive functionality.

        Example:
            >>> from vectorai.client import ViClient
            >>> vi_client = ViClient(username, api_key, vectorai_url)
            >>> doc = {'items': {'chicken': 'fried'}, 'food_vector_': [0, 1, 2]}
            >>> vi_client._check_schema(doc)
        """
        VECTOR_FIELD_NAME = "_vector_"
        IS_VECTOR_FIELD_MISSING = True
        IS_ID_FIELD_MISSING = True
        for field, value in document.items():
            if field == '_id':
                IS_ID_FIELD_MISSING = False
            if isinstance(value, dict):
                IS_ID_FIELD_MISSING, IS_VECTOR_FIELD_MISSING = self._check_schema(
                    document[field],
                    is_missing_vector_field=IS_VECTOR_FIELD_MISSING,
                    is_missing_id_field=IS_ID_FIELD_MISSING,
                    is_nested=True
                )
            if "_vectors_" in field:
                warnings.warn(
                    "Rename " + field + "to " + field.replace('_vectors_', '_vector_')
                , MissingFieldWarning)
        
        for field in document.keys():
            if VECTOR_FIELD_NAME in field:
                IS_VECTOR_FIELD_MISSING = False
        
        if not is_nested:
            if IS_VECTOR_FIELD_MISSING:
                warnings.warn(
                    "Potential issue. Cannot find a vector field. Check that the vector field contains _vector_.",
                    MissingFieldWarning
                )
            if IS_ID_FIELD_MISSING:
                warnings.warn(
                    "Missing ID field. Please include an _id field to make inserting easier.",
                    MissingFieldWarning
                )
        return IS_ID_FIELD_MISSING, IS_VECTOR_FIELD_MISSING

    def list_collections(self) -> List[str]:
        """
        List Collections

        Args:
            username:
		        Username
	        api_key:
		        Api Key, you can request it from request_api_key

        Returns:
            List of collections

        Example:
            >>> from vectorai.client import ViClient
            >>> vi_client = ViClient(username, api_key, vectorai_url)
            >>> doc = {'items': {'chicken': 'fried'}, 'food_vector_': [0, 1, 2]}
            >>> vi_client._check_schema(doc)
        """
        return sorted(self._list_collections())

    def create_filter_query(self, collection_name: str, field: str, filter_type: str, filter_values: Union[List[str], str]=None):
        """
            Filter type can be one of contains/exact_match/categories/exists/insert_date/numeric_range
            Filter types can be one of:
            contains: Field must contain this specific string. Not case sensitive.
            exact_match: Field must have an exact match 
            categories: Matches entire field
            exists: If field exists in document 
            >= / > / < / <= : Larger than or equal to / Larger than / Smaller than / Smaller than or equal to
            These, however, can only be applied on numeric/date values. Check collection_schema.

            Args:
            collection_name: The name of the collection 
            field: The field to filter on 
            filter_type: One of contains/exact_match/categories/>=/>/<=/<.

        """
        print("After creating query, you can now apply this to filters.")
        if filter_type == 'contains':
            return [{'field' : field, 'filter_type' : 'contains', "condition":"==", "condition_value": filter_values}]
        if filter_type == 'exact_match':
            return [{'field' : field, 'filter_type' : 'exact_match', "condition":"==", "condition_value": filter_values}]
        if filter_type == 'categories':
            return [{'field' : field, 'filter_type' : 'categories', "condition":"==", "condition_value": filter_values}]
        if filter_type == 'exists':
            return [{'field' : field, 'filter_type' : 'exists', "condition":">=", "condition_value":" "}]
        if filter_type == '<=' or filter_type == '>=' or filter_type == '>' or filter_type == '<' or filter_type == '==':
            if self.collection_schema(collection_name)[field] == 'date':
                return [{'field' : field, 'filter_type' : 'date', "condition":filter_type, "condition_value": filter_values}]
            elif self.collection_schema(collection_name)[field] == 'numeric':
                return [{'field' : field, 'filter_type' : 'numeric', "condition":filter_type, "condition_value":filter_values}]
        else:
            raise ValueError(f"{filter_type} has not been defined. Please choose one of contains/exact_match/exists/categories/>=/<=/>/<.")



"""Errors
"""


class APIError(Exception):
    """Base error class for all errors in library
    """

    def __init__(self, response_message: str=None):
        """
        The main Vi  base error.

        Args:
            response_message: THe error message

        Example:
            >>> raise APIError("Missing ____.")
        """
        self.response_message = response_message

class MissingFieldWarning(APIError, UserWarning):
    """
        Warning for missing field. Used for checking collection schema
        upon insertion.
    """
    pass

class MissingFieldError(APIError):
    """
        Error in case the field is missing from a document.
        Used for when a specific field is missing.
    """
    pass

class LoginError(APIError):
    """
    Login Error
    """
    pass



import requests
from typing import Dict, List
from .read import ViReadAPIClient
from .utils import retry

class ViWriteAPIClient(ViReadAPIClient):
    """
    Write Operations
    """
    def __init__(self, username, api_key, url=None):
        self.username = username
        self.api_key = api_key
        if url:
            self.url = url
        else:
            self.url = "https://api.vctr.ai"

    def create_collection_from_document(self, collection_name: str, document: dict):
        """
Creates a collection by infering the schema from a document

If you are inserting your own vector use the suffix (ends with)  **"\_vector\_"** for the field name. e.g. "product\_description\_vector\_"
    
Args:
	collection_name:
		Name of Collection
	document:
		A Document is a JSON-like data that we store our metadata and vectors with. For specifying id of the document use the field '\_id', for specifying vector field use the suffix of '\_vector\_'
"""
        return requests.post(
            url="{}/project/create_collection_from_document".format(self.url),
            json={
                "username": self.username,
                "api_key": self.api_key,
                "collection_name": collection_name,
                "document": document,
            },
        ).json()

    @retry()
    def bulk_insert_and_encode(self, collection_name: str, docs: list, models: dict):
        """
            Client-side encoding of documents to improve speed of inserting. This removes
            the step of retrieving the vectors and can be useful to accelerate the encoding
            process if required.
            Models can be one of 'text', 'audio' or 'image'.
        """
        # For each of the field values in the models, check that the deployed models
        # are in the list below.
        for k, v in models:
            assert v in ['text', 'audio', 'image']
        return requests.post(
            url='{}/collection/bulk_insert_and_encode'.format(url), 
            json={
            'username' : self.username,
            'api_key' : self.api_key,
            'collection_name': collection_name,
            'documents' : docs,
            'models' : models
        }).json()

    def _create_collection(self, collection_name: str, collection_schema: Dict = {}):
        return requests.post(
            url="{}/project/create_collection".format(self.url),
            json={
                "username": self.username,
                "api_key": self.api_key,
                "collection_name": collection_name,
                "collection_schema": collection_schema,
            },
        ).json()

    def _delete_collection(self, collection_name: str):
        return requests.get(
            url="{}/project/delete_collection".format(self.url),
            params={
                "username": self.username,
                "api_key": self.api_key,
                "collection_name": collection_name,
            },
        ).json()

    # def replicate_collection(self, collection_name: str, new_collection_name: str):
    #     return requests.get(
    #         url="{}/project/replicate_collection".format(self.url),
    #         params={
    #             "username": self.username,
    #             "api_key": self.api_key,
    #             "collection_name": collection_name,
    #             "new_collection_name": collection_name,
    #         },
    #     ).json()

    @retry()
    def bulk_insert(self, collection_name: str, documents: List, insert_date: bool=True, overwrite: bool=True):
        """
Insert multiple documents into a Collection
When inserting the document you can specify your own id for a document by using the field name **"\_id"**. 
For specifying your own vector use the suffix (ends with)  **"\_vector\_"** for the field name.
e.g. "product\_description\_vector\_"
    
Args:
	collection_name:
		Name of Collection
	documents:
		A list of documents. Document is a JSON-like data that we store our metadata and vectors with. For specifying id of the document use the field '\_id', for specifying vector field use the suffix of '\_vector\_'
	insert_date:
		Whether to include insert date as a field 'insert_date_'.
	overwrite:
		Whether to overwrite document if it exists.
"""
        return requests.post(
            url="{}/collection/bulk_insert".format(self.url),
            json={
                "username": self.username,
                "api_key": self.api_key,
                "collection_name": collection_name,
                "documents": documents,
                "insert_date": insert_date,
                "overwrite" : overwrite
            },
        ).json()
    
    @retry()
    def insert(self, collection_name: str, document: Dict, insert_date: bool=True, overwrite: bool=True):
        """
Insert a document into a Collection
When inserting the document you can specify your own id for a document by using the field name **"\_id"**. 
For specifying your own vector use the suffix (ends with)  **"\_vector\_"** for the field name.
e.g. "product\_description\_vector\_"
    
Args:
	collection_name:
		Name of Collection
	document:
		A Document is a JSON-like data that we store our metadata and vectors with. For specifying id of the document use the field '\_id', for specifying vector field use the suffix of '\_vector\_'
	insert_date:
		Whether to include insert date as a field 'insert_date_'.
	overwrite:
		Whether to overwrite document if it exists.
"""
        return requests.post(
            url="{}/collection/insert".format(self.url),
            json={
                "username": self.username,
                "api_key": self.api_key,
                "collection_name": collection_name,
                "document": document,
                "insert_date": insert_date,
                "overwrite" : overwrite
            },
        ).json()
    
    @retry()
    def _edit_document(self, collection_name: str, edits: Dict, document_id: str):
        return requests.post(
            url="{}/collection/edit_document".format(self.url),
            json={
                "username": self.username,
                "api_key": self.api_key,
                "collection_name": collection_name,
                "edits": edits,
                "document_id": document_id,
            },
        ).json()

    def delete_by_id(self, collection_name: str, document_id: str):
        """
Delete a document in a Collection by its id
    
Args:
	document_id:
		ID of a document
	collection_name:
		Name of Collection
"""
        return requests.get(
            url="{}/collection/delete_by_id".format(self.url),
            params={
                "username": self.username,
                "api_key": self.api_key,
                "collection_name": collection_name,
                "document_id": document_id,
            },
        ).json()

    def publish_aggregation(
        self,
        collection_name: str,
        aggregation_query: dict,
        aggregation_name: str,
        aggregated_collection_name: str,
        description: str = "published aggregation",
        date_field: str = "insert_date_",
        refresh_time: int = 30,
        start_immediately: bool = True,
    ):
        """
Publishes your aggregation query to a new collection
Publish and schedules your aggregation query and saves it to a new collection.
This new collection is just like any other collection and you can read, filter and aggregate it.
    
Args:
	source_collection:
		The collection where the data to aggregate comes from
	dest_collection:
		The name of collection of where the data will be aggregated to
	aggregation_name:
		The name for the published scheduled aggregation
	description:
		The description for the published scheduled aggregation
	aggregation_query:
		The aggregation query to schedule
	date_field:
		The date field to check whether there is new data coming in
	refresh_time:
		How often should the aggregation check for new data
	start_immediately:
		Whether to start the published aggregation immediately
"""
        return requests.post(
            url="{}/collection/publish_aggregation".format(self.url),
            json={
                "username": self.username,
                "api_key": self.api_key,
                "aggregation_query": aggregation_query,
                "source_collection": collection_name,
                "dest_collection": aggregated_collection_name,
                "aggregation_name": aggregation_name,
                "description": description,
                "date_field": date_field,
                "refresh_time": "{}s".format(refresh_time),
                "start_immediately": start_immediately,
            },
        ).json()

    def start_aggregation(self, aggregation_name: str):
        """
Start your published aggregation
Start or resume your published aggregation. The published aggregation can be stopped with /stop_aggregation.
    
Args:
	aggregation_name:
		The name for the published scheduled aggregation
"""
        return requests.get(
            url="{}/collection/start_aggregation".format(self.url),
            params={
                "username": self.username,
                "api_key": self.api_key,
                "aggregation_name": aggregation_name,
            },
        ).json()

    def stop_aggregation(self, aggregation_name: str):
        """
Stop your published aggregation
Stop/pause your published aggregation. The published aggregation can be resumed/started with /start_aggregation.
    
Args:
	aggregation_name:
		The name for the published scheduled aggregation
"""
        return requests.get(
            url="{}/collection/stop_aggregation".format(self.url),
            params={
                "username": self.username,
                "api_key": self.api_key,
                "aggregation_name": aggregation_name,
            },
        ).json()

    def delete_published_aggregation(self, aggregation_name: str):
        """
Delete a published aggregation and collection
Delete a published aggregation and the associated collection it creates.
    
Args:
	aggregation_name:
		The name for the published scheduled aggregation
"""
        return requests.get(
            url="{}/collection/delete_published_aggregation".format(self.url),
            params={
                "username": self.username,
                "api_key": self.api_key,
                "aggregation_name": aggregation_name,
            },
        ).json()

    def join_collections(self, join_query: dict, joined_collection_name:str):
        """
Join collections with a query
Perform a join query on a whole collection and write the results to a new collection. We currently only support left joins.
    
Args:
	join_query:
		.
	joined_collection_name:
		Name of the new collection that contains the joined results
"""
        return requests.post(
            url="{}/collection/join_collections".format(self.url),
            json={
                "username": self.username,
                "api_key": self.api_key,
                "join_query": join_query,
                "joined_collection_name": joined_collection_name,
            },
        ).json()


from .search import *
from .read import *
from .write import *
from .text import *
from .audio import *
from .text import *
from .cluster import *
from .dimensionality_reduction import *
from .array_dict_vectorizer import *


import io
import base64
import requests
from typing import Dict, List


class ViArrayDictClient:
    """
    Search and Encoding for Array & Dictionary
    """
    def __init__(self, username, api_key, url=None):
        self.username = username
        self.api_key = api_key
        if url:
            self.url = url
        else:
            self.url = "https://api.vctr.ai"


    def encode_dictionary_field(self, collection_name: str, dictionary_fields: List):
        """
Encode all dictionaries in a field for collection into vectors

Within a collection encode the specified dictionary field in every document into vectors.

For example: a dictionary that represents a **person's characteristics visiting a store, field "person_characteristics"**::

    document 1 field: {"person_characteristics" : {"height":180, "age":40, "weight":70}}

    document 2 field: {"person_characteristics" : {"age":32, "purchases":10, "visits": 24}}

    -> <Encode the dictionaries to vectors> ->

    | height | age | weight | purchases | visits |
    |--------|-----|--------|-----------|--------|
    | 180    | 40  | 70     | 0         | 0      |
    | 0      | 32  | 0      | 10        | 24     |

    document 1 dictionary vector: {"person_characteristics_vector_": [180, 40, 70, 0, 0]}

    document 2 dictionary vector: {"person_characteristics_vector_": [0, 32, 0, 10, 24]}

Args:
    dictionary_fields:
        The dictionary field to train on to encode into vectors
    collection_name:
        Name of Collection
"""
        return requests.get(
            url="{}/collection/encode_dictionary_field".format(self.url),
            params={
                "username": self.username,
                "api_key": self.api_key,
                "collection_name": collection_name,
                "dictionary_fields": dictionary_fields,
            },
        ).json()

    def encode_dictionary(self, collection_name: str, dictionary: Dict, dictionary_field: str):
        """
Encode an dictionary into a vector

For example: a dictionary that represents a **person's characteristics visiting a store, field "person_characteristics"**::

    {"height":180, "age":40, "weight":70}

    -> <Encode the dictionary to vector> ->

    | height | age | weight | purchases | visits |
    |--------|-----|--------|-----------|--------|
    | 180    | 40  | 70     | 0         | 0      |

    dictionary vector: [180, 40, 70, 0, 0]

Args:
	collection_name:
		Name of Collection
	dictionary:
		A dictionary to encode into vectors
	dictionary_field:
		The dictionary field that encoding of the dictionary is trained on
"""
        return requests.post(
            url="{}/collection/encode_dictionary".format(self.url),
            json={
                "username": self.username,
                "api_key": self.api_key,
                "collection_name": collection_name,
                "dictionary": dictionary,
                "dictionary_field": dictionary_field,
            },
        ).json()

    def search_with_dictionary(self, collection_name: str, dictionary: Dict, dictionary_field: str,
        fields: List,
        sum_fields: bool = True,
        metric: str = "cosine",
        min_score=None,
        page: int = 1,
        page_size: int = 10,
        include_vector:bool=False,
        include_count:bool=True,
        asc:bool=False):
        """
Search a dictionary field with a dictionary using Vector Search with a dictionary directly.

For example: a dictionary that represents a **person's characteristics visiting a store, field "person_characteristics"**::

    {"height":180, "age":40, "weight":70}

    -> <Encode the dictionary to vector> ->

    | height | age | weight | purchases | visits |
    |--------|-----|--------|-----------|--------|
    | 180    | 40  | 70     | 0         | 0      |

    dictionary vector: [180, 40, 70, 0, 0]

    -> <Vector Search> ->

    Search Results: {...}

Args:
	collection_name:
		Name of Collection
	search_fields:
		Vector fields to search against
	page_size:
		Size of each page of results
	page:
		Page of the results
	approx:
		Used for approximate search
	sum_fields:
		Whether to sum the multiple vectors similarity search score as 1 or seperate
	metric:
		Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
	min_score:
		Minimum score for similarity metric
	include_vector:
		Include vectors in the search results
	include_count:
		Include count in the search results
	hundred_scale:
		Whether to scale up the metric by 100
	dictionary:
		A dictionary to encode into vectors
	dictionary_field:
		The dictionary field that encoding of the dictionary is trained on
    asc:
        Whether to sort the score by ascending order (default is false, for getting most similar results)
"""
        return requests.post(
            url="{}/collection/search_with_dictionary".format(self.url),
            json={
                "username": self.username,
                "api_key": self.api_key,
                "collection_name": collection_name,
                "dictionary": dictionary,
                "dictionary_field": dictionary_field,
                "sum_fields": sum_fields,
                "search_fields": fields,
                "metric": metric,
                "min_score": min_score,
                "page": page,
                "page_size": page_size,
                "include_vector": include_vector,
                "include_count": include_count,
                "asc": asc,
            },
        ).json()

    def encode_array_field(self, collection_name: str, array_fields: List):
        """
Encode all arrays in a field for a collection into vectors

Within a collection encode the specified array field in every document into vectors.

For example, array that represents a ****movie's categories, field "movie_categories"**::

    document 1 array field: {"category" : ["sci-fi", "thriller", "comedy"]}

    document 2 array field: {"category" : ["sci-fi", "romance", "drama"]}

    -> <Encode the arrays to vectors> ->

    | sci-fi | thriller | comedy | romance | drama |
    |--------|----------|--------|---------|-------|
    | 1      | 1        | 1      | 0       | 0     |
    | 1      | 0        | 0      | 1       | 1     |

    document 1 array vector: {"movie_categories_vector_": [1, 1, 1, 0, 0]}

    document 2 array vector: {"movie_categories_vector_": [1, 0, 0, 1, 1]}

Args:
	array_fields:
		The array field to train on to encode into vectors
	collection_name:
		Name of Collection
"""
        return requests.get(
            url="{}/collection/encode_array_field".format(self.url),
            params={
                "username": self.username,
                "api_key": self.api_key,
                "collection_name": collection_name,
                "array_fields": array_fields,
            },
        ).json()

    def encode_array(self, collection_name: str, array: List, array_field: str):
        """
Encode an array into a vector

For example: an array that represents a **movie's categories, field "movie_categories"**::

    ["sci-fi", "thriller", "comedy"]

    -> <Encode the arrays to vectors> ->

    | sci-fi | thriller | comedy | romance | drama |
    |--------|----------|--------|---------|-------|
    | 1      | 1        | 1      | 0       | 0     |

    array vector: [1, 1, 1, 0, 0]

Args:
	array_field:
		The array field that encoding of the dictionary is trained on
	array:
		The array to encode into vectors
	collection_name:
		Name of Collection
"""
        return requests.get(
            url="{}/collection/encode_array".format(self.url),
            params={
                "username": self.username,
                "api_key": self.api_key,
                "collection_name": collection_name,
                "array": array,
                "array_field": array_field,
            },
        ).json()

    def search_with_array(self, collection_name: str, array: List, array_field: str,
        fields: List,
        sum_fields: bool = True,
        metric: str = "cosine",
        min_score=None,
        page: int = 1,
        page_size: int = 10,
        include_vector:bool=False,
        include_count:bool=True,
        asc:bool=False):
        """
Search an array field with an array using Vector Search with an array directly.

For example: an array that represents a **movie's categories, field "movie_categories"**::

    ["sci-fi", "thriller", "comedy"]

    -> <Encode the arrays to vectors> ->

    | sci-fi | thriller | comedy | romance | drama |
    |--------|----------|--------|---------|-------|
    | 1      | 1        | 1      | 0       | 0     |

    array vector: [1, 1, 1, 0, 0]

    -> <Vector Search> ->

    Search Results: {...}
    
Args:
	array_field:
		The array field that encoding of the dictionary is trained on
	array:
		The array to encode into vectors
	collection_name:
		Name of Collection
	search_fields:
		Vector fields to search through
	approx:
		Used for approximate search
	sum_fields:
		Whether to sum the multiple vectors similarity search score as 1 or seperate
	page_size:
		Size of each page of results
	page:
		Page of the results
	metric:
		Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
	min_score:
		Minimum score for similarity metric
	include_vector:
		Include vectors in the search results
	include_count:
		Include count in the search results
	hundred_scale:
		Whether to scale up the metric by 100
    asc:
        Whether to sort the score by ascending order (default is false, for getting most similar results)
"""
        return requests.get(
            url="{}/collection/search_with_array".format(self.url),
            params={
                "username": self.username,
                "api_key": self.api_key,
                "collection_name": collection_name,
                "array": array,
                "array_field": array_field,
                "sum_fields": sum_fields,
                "search_fields": fields,
                "metric": metric,
                "min_score": min_score,
                "page": page,
                "page_size": page_size,
                "include_vector": include_vector,
                "include_count": include_count,
                "asc": asc,
            },
        ).json()



"""
    Utilities for relational documents.
"""
from ..models.deployed.base import ViDeployedModel
import numpy as np
from itertools import permutations
from typing import List, Dict

def vector_operation(vector_1: list, vector_2: list, operation: str='mean', axis: int=0):
    """
        Perform vector operation between 2 vectors.
        Args:
            vector_1: First vector
            vector_2: Second vector
            Operation: The operation that can be used (mean/minus/sum/min/max)
            Axis: Axis on which to perform the vector operation
        Example:
            >>> vector_operation(np.array([1, 2, 3]), np.array([2, 3, 4]))
    """
    vectors = [vector_1, vector_2]
    if operation == "mean":
        return np.mean(vectors, axis=axis).tolist()
    elif operation == 'minus':
        return np.subtract(vector_1, vector_2).tolist()
    elif operation == "sum":
        return np.sum(vectors, axis=axis).tolist()
    elif operation == "min":
        return np.min(vectors, axis=axis).tolist()
    elif operation == "max":
        return np.max(vectors, axis=axis).tolist()
    else:
        raise ValueError("Vector operation needs to be one of minus/mean/sum/min/max.")

def create_relational_document(doc_1: Dict, doc_2: Dict, vector_fields: List[str], label_field: str, operation="minus"):
    """
        Args:
            Operation: minus/sum
            Vector_fields: The name of the vector fields
            label: The entity of the vector
        Example:
            >>> from vectorai.utils import UtilsMixin
            >>> mixin_utils = UtilsMixin()
            >>> doc = mixin_utils.create_sample_document()
            >>> doc_2 = mixin_utils.create_sample_document()
            >>> create_relational_document(doc_1, doc_2, vector_fields=['color_vector_'], label_field='country')
    """
    # Create a new document
    new_doc = {}
    for vector in vector_fields:
        new_vector = vector_operation(doc_1[vector], doc_2[vector], operation=operation)
        if operation.lower() == 'minus':
            new_label = doc_1[label_field] + " - " + doc_2[label_field]
        elif operation.lower() == 'sum':
            new_label = doc_1[label_field] + " + " + doc_2[label_field]
        elif operation.lower() == 'mean':
            new_label = "avg(" + doc_1[label_field] + ", " + doc_2[label_field] + ")"
        elif operation.lower() == 'max': 
            new_label = "max(" + doc_1[label_field] + ", " + doc_2[label_field] + ")"
        elif operation.lower() == 'min':
            new_label = "min(" + doc_1[label_field] + ", " + doc_2[label_field] + ")"
        new_doc.update({label_field: new_label, vector : new_vector})
    new_doc.update({'doc_1': doc_1, 'doc_2': doc_2})
    return new_doc

def create_relational_collection(docs: list, label_field: str, vector_fields: List[str], operation: str='minus'):
    """
        Create relational collection that.
        It should return n*(n-1) documents (for every permutation of the documents.
        Args:
            docs: The documents
            label_field: The field of the labels
            vector_fields: The vector fields in the document 
            operation: Must be one of minus/mean/max/min/sum.
        Example:
            create_relational_collection(docs)
    """
    perm = list(permutations(docs, 2))
    print("Returning n * (n - 1) documents. Insert and search to explore what relationships you end up with!")
    return [create_relational_document(*c, vector_fields=vector_fields, operation=operation, label_field=label_field) for i, c in enumerate(perm)]



import io
import base64
import requests
import numpy as np
from typing import List
from .base import ViDeployedModel


class ViText2Vec(ViDeployedModel):
    def encode(self, text: str):
        """
            Convert text to vectors.
        """
        return requests.get(
            url="{}/collection/encode_text".format(self.url),
            params={
                "username": self.username,
                "api_key": self.api_key,
                "collection_name": self.collection_name,
                "text": text,
            },
        ).json()

    def bulk_encode(self, texts: List[str]):
        """
            Bulk convert text to vectors
        """
        return requests.get(
            url="{}/collection/bulk_encode_text".format(self.url),
            params={
                "username": self.username,
                "api_key": self.api_key,
                "collection_name": self.collection_name,
                "texts": texts,
            }
        ).json()

    @property
    def __name__(self):
        if self._name is None:
            return "vectorai_text"
        return self._name

    @__name__.setter
    def __name__(self, value):
        self._name = value


class ViTextArray2Vec(ViDeployedModel):
    def __init__(
        self,
        username,
        api_key,
        url=None,
        collection_name="base",
        vector_operation: str = "mean",
    ):
        self.username = username
        self.api_key = api_key
        if url:
            self.url = url
        else:
            self.url = "https://api.vctr.ai"
        self.collection_name = collection_name
        self.vector_operation = vector_operation

    def encode(self, texts):
        return self._vector_operation(
            requests.get(
                url="{}/collection/bulk_encode_text".format(self.url),
                params={
                    "username": self.username,
                    "api_key": self.api_key,
                    "collection_name": self.collection_name,
                    "texts": texts,
                }
            ).json(),
            vector_operation=self.vector_operation,
        )

    @property
    def __name__(self):
        if self._name is None:
            return "vectorai_text_array"
        return self._name

    @__name__.setter
    def __name__(self, value):
        self._name = value



from typing import List
import requests


class ViDimensionalityReductionClient:
    """
    Dimensionality Reduction
    """
    def __init__(self, username: str, api_key: str, url=None):
        self.username = username
        self.api_key = api_key
        if url:
            self.url = url
        else:
            self.url = "https://api.vctr.ai"

    def dimensionality_reduce(
        self,
        collection_name: str,
        vectors: List[List[float]],
        vector_field: str,
        n_components: int,
        alias: str = "default",
    ):
        """
Trains a Dimensionality Reduction model on the collection

Dimensionality reduction allows your vectors to be reduced down to any dimensions greater than 0 using unsupervised machine learning. This is useful for even faster search and visualising the vectors.
        
Args:
	vector_field:
		Vector field to perform dimensionality reduction on
	alias:
		Alias is used to name the dimensionality reduced vectors
	n_components:
		The size/length to reduce the vector down to. If 0 is set then highest possible is of components is set, when this is done you can get reduction on demand of any length.
	refresh:
		Whether to refresh the whole collection and retrain the dimensionality reduction model
	collection_name:
		Name of Collection
"""
        return requests.get(
            url="{}/collection/dimensionality_reduce".format(self.url),
            params={
                "username": self.username,
                "api_key": self.api_key,
                "collection_name": collection_name,
                "vectors": vectors,
                "vector_field": vector_field,
                "alias": alias,
                "n_components": n_components,
            },
        ).json()

    def dimensionality_reduction_job(
        self,
        collection_name: str,
        vector_field: str,
        n_components: int = 0,
        alias: str = "default",
        refresh: bool = True,
    ):
        """
Trains a Dimensionality Reduction model on the collection

Dimensionality reduction allows your vectors to be reduced down to any dimensions greater than 0 using unsupervised machine learning. This is useful for even faster search and visualising the vectors.
        
Args:
	vector_field:
		Vector field to perform dimensionality reduction on
	alias:
		Alias is used to name the dimensionality reduced vectors
	n_components:
		The size/length to reduce the vector down to. If 0 is set then highest possible is of components is set, when this is done you can get reduction on demand of any length.
	refresh:
		Whether to refresh the whole collection and retrain the dimensionality reduction model
	collection_name:
		Name of Collection
"""
        return requests.get(
            url="{}/collection/jobs/dimensionality_reduction".format(self.url),
            params={
                "username": self.username,
                "api_key": self.api_key,
                "collection_name": collection_name,
                "vector_field": vector_field,
                "alias": alias,
                "n_components": n_components,
                "refresh": refresh,
            },
        ).json()



"""
    Add default retry to inserting
"""
import time
from functools import wraps

def retry(num_of_retries=3, timeout=2):
    """
    Allows the function to retry upon failure. 
    Args:
        num_of_retries: The number of times the function should retry
        timeout: The number of seconds to wait between each retry
    """
    def _retry(func):
        @wraps(func)
        def function_wrapper(*args, **kwargs):
            for i in range(num_of_retries):
                try:
                    return func(*args, **kwargs)
                # Using general error to avoid any possible error dependencies.
                except Exception as error:
                    time.sleep(timeout)
                    print("Retrying...")
                    if i == num_of_retries - 1:
                        raise error
                    continue
                break
        return function_wrapper
    return _retry



"""
Plotting functions of Vi go here. 
"""
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from tqdm.notebook import tqdm
import copy
from typing import Union, List, Dict, Any
from .score import ViScore
from .utils import ViAnalyticsUtils, MeanDict
from ..read import ViReadClient
from ..errors import APIError

class VizMixin(ViScore, ViAnalyticsUtils):
    """
    Visualisation submodule for the library.
    """

    def _scatter_plot_documents(
        self,
        filtered_collection: list,
        dim_reduction_field: str,
        color: str,
        point_label: str,
        title: str=None,
        mode: str='markers',
        marker_size:int = 5
    ):
        """
        Add additional plot to figure and filtered collection and dim reduction field.

        Args:
            filtered_collection:
                A collection or the name of a collection
            dim_reduction_field:
                Dimensionally reduced field.
            color:
                Color of the plot
            point_label:
                Label of the plot
            title:
                Name of the group of documents

        Returns:
            go.Scatter: Scatter plot of centroids.

        """
        return go.Scatter(
            x=[
                self.get_field(dim_reduction_field, x)[0]
                for x in filtered_collection
            ],
            y=[
                self.get_field(dim_reduction_field, x)[1]
                for x in filtered_collection
            ],
            mode=mode,
            marker=dict(size=marker_size, color=color),
            text=[x[point_label] for x in filtered_collection],
            name=title,
        )

    def _scatter_plot_centroids(
        self,
        cluster_centroid_documents: List[Dict],
        dim_reduction_field: str,
        point_label: str,
        line_color="DarkSlateGrey",
        line_width=2,
        size=5,
        color_dict=None,
        mode: str='markers',
        color='orange',
    ):
        """
        Scatterplot the centroids

        Args:
            cluster_centroid_documents:
                The centroids to highlight.
            point_label:
                The label of every point. This should be found in the document.
            dim_reduction_field:
                The dimensionally-reduced vectors.
            line_width:
                The width of the lines.
            line_color:
                THe line color of the centroids
            color_dict:
                What to color the centroid the cluster label/title
            size:
                The size of the markers for the centroid documents
        
        Returns:
            go.Scatter: Scatter plot of the documents
        
        """
        x = [
            self.get_field(dim_reduction_field, x)[0]
            for x in cluster_centroid_documents.values()
        ]
        y = [
            self.get_field(dim_reduction_field, x)[1]
            for x in cluster_centroid_documents.values()
        ]
        titles = [
            self.get_field(point_label, x)
            for x in cluster_centroid_documents.values()
        ]
        return go.Scatter(
            x=x,
            y=y,
            mode=mode,
            marker=dict(
                size=size,
                # color=[color_dict[str(x['_clusters_'][cluster_field]['title'])] for x in cluster_centroids.values()],
                color=color,
                line=dict(width=line_width, color=line_color),
            ),
            text=titles,
            name="centroids",
        )

    def plot_dimensionality_reduced_vectors(
        self,
        collection: Union[str, List[Dict]],
        point_label: str,
        dim_reduction_field: str,
        cluster_field: str=None,
        cluster_label: str=None,
        include_centroids: bool = False,
        color: str=None,
        alias: str = None,
        mode: str='markers',
    ):
        """
        Returns a 2D plot of vectors that have been dimensionally reduced.

        Args:
            collection:
                A collection or the name of a collection
            point_label:
                The label of every point. This should be found in the document.
            dim_reduction_field:
                The dimensionally-reduced vectors.
            cluster_field:
                The field by which it is clustered.
            cluster_label:
                The name of the clusters
            include_centroids:
                Whether to include the centroids of every cluster
        
        Returns:
            Plotly figure object
        
        See example from: https://colab.research.google.com/drive/10u7b3lkIVJ-lceCmr34ywscOGueIFd0I?usp=sharing
        
        Example:
            >>> collection_name = 'nlp-qa'
            >>> cluster_field = 'question_vector_'
            >>> cluster_label = 'category'
            >>> alias = '1st_cluster'
            >>> dim_reduction_field = '_dr_.default.2.question_vectors_' # the '.' in names implies nested dictionaries in Vi
            >>> vi_client.plot_dimensionality_reduced_vectors(collection=collection_name,
                    cluster_field=cluster_field,
                    cluster_label=cluster_label,
                    point_label='question_title',
                    dim_reduction_field=dim_reduction_field, 
                    include_centroids=True, 
                    alias=alias)
        
        """
        fig = go.Figure()
        if include_centroids:
            if not isinstance(collection, str):
                raise NotImplementedError(
                    "Centroid visualisation not included for custom clustering."
                    + "Please add them to yourself by using add_trace to the returned fig output."
                )

            cluster_centroid_documents = self.advanced_cluster_centroid_documents(
                collection, cluster_field, alias=alias
            )

            fig.add_trace(
                self._scatter_plot_centroids(
                    cluster_centroid_documents=cluster_centroid_documents,
                    dim_reduction_field=dim_reduction_field,
                    point_label=point_label,
                    mode=mode
                )
            )

        # If collection is simply a name, retrieve collection
        if isinstance(collection, str):
            fields_to_include = [point_label, dim_reduction_field, cluster_field, cluster_label]
            fields_to_include = [x for x in fields_to_include if x is not None]
            collection = self.retrieve_all_documents(collection, include_fields=[fields_to_include])

        if cluster_label is None:
            color_dict = None
        else:
            cluster_titles = [
                self.get_field(cluster_label, x) for x in collection
            ]
            cluster_titles = list(set(cluster_titles))
            cluster_titles_color_map = dict(enumerate(cluster_titles))
            color_dict = {v: k + 5 for k, v in cluster_titles_color_map.items()}
        
        if cluster_label is None:
            fig.add_trace(
                self._scatter_plot_documents(
                    filtered_collection=collection,
                    dim_reduction_field=dim_reduction_field,
                    color=color,
                    point_label=point_label,
                    mode=mode
                )
            )
        else:
            for title in tqdm(cluster_titles):
                filtered_collection = [
                    x
                    for x in collection
                    if self.get_field(cluster_label, x) == title
                ]
                fig.add_trace(
                    self._scatter_plot_documents(
                        title=title,
                        filtered_collection=filtered_collection,
                        dim_reduction_field=dim_reduction_field,
                        color=color_dict[title],
                        point_label=point_label,
                    )
                )
            fig.update_layout(showlegend=True)
        return fig

    def plot_1d_cosine_similarity(
        self,
        documents: list,
        vector_fields: List[str],
        label: str,
        anchor_document: dict = None,
        anchor_index: Union[str, int] = 0,
        orientation: int="h",
        barmode: int="group",
        num_cols: int=None,
        y_axis_tickangle: int=-15,
        x_axis_tickangle: int=15
    ):
        """
        Compare 1 document against other documents.
        Ensure that the name is unique, otherwise the plot will simply take the mean.
        
        Args:
            documents: list of documents (dictionaries) to feed in
            vector_fields: vector field to calculate cosine similarity on 
            label: the x label for the bar plot 
            anchor_document: the document to compare it on
            anchor_index: the anchor index to compare it on
            orientation: The orientation of the bar chart. Can be 'v' (vertical) or 'h' (horizontal)
            num_cols: The number of columns. The default will put everything into 1 row. If you want to 
            put things into multiple rows, then please reduce the number of columns. 
            y_axis_tickangle: This will change the tick angles of the y axis in the vertical chart.
            x_axis_tickangle: This will change the tick angels of the x axis in the vertical chart.
        Returns:
            Plotly Figure: returns a horizontal barplot showing cosine similarity scores.
        
        Example:
            >>> cluster_field = 'question_vector_'
            >>> cluster_label = 'category'
            >>> alias = '1st_cluster'
            >>> dim_reduction_field = '_dr_.default.2.question_vectors_' # the '.' in names implies nested dictionaries in Vi
            >>> vi_client.plot_1d_cosine_similarity(collection=collection_name,
                    cluster_field=cluster_field,
                    cluster_label=cluster_label,
                    point_label='question_title',
                    dim_reduction_field=dim_reduction_field, 
                    include_centroids=True, 
                    alias=alias)
        
        """
        if isinstance(vector_fields, str):
            vector_fields = [vector_fields]
        if anchor_document is None:
            other_documents = copy.deepcopy(documents)
            if isinstance(anchor_index, int):
                anchor_document = other_documents.pop(anchor_index)
            elif isinstance(anchor_index, str):
                # Loop through list to get ID
                for i, doc in enumerate(other_documents):
                    if anchor_index == doc["_id"]:
                        anchor_document = other_documents.pop(i)
                        break
        import math
        if num_cols is None:
            num_cols = len(vector_fields)
        fig = make_subplots(rows=int(len(vector_fields) / num_cols) + 1, cols=num_cols, 
        subplot_titles=vector_fields)
        row_num = 1
        col_num = 1
        for i, vector_field in enumerate(vector_fields):
            fig.add_trace(self._plot_1d_cosine_similarity_for_1_vector_field(
                documents=documents,
                vector_field=vector_field,
                label=label,
                anchor_document=anchor_document, 
                anchor_index=anchor_index,
                orientation=orientation),
                row=row_num, col=col_num)
            if col_num % num_cols == 0:
                row_num += 1
                col_num = 1
            else:
                col_num += 1
        if orientation == 'h':
            fig.update_yaxes(tickangle=y_axis_tickangle)
        if orientation == 'v':
            fig.update_xaxes(tickangle=x_axis_tickangle)
        
        return fig
    
    def _plot_1d_cosine_similarity_for_1_vector_field(self, 
        documents: list,
        vector_field: str,
        label: str,
        anchor_document: dict=None,
        anchor_index: Union[str, int]=0,
        orientation="h") -> go.Bar:
            """
            Compare 1 document against other documents.

            Args:
                documents: list of documents (dictionaries) to feed in
                vector_fields: vector field to calculate cosine similarity on 
                label: the x label for the bar plot 
                anchor_document: the document to compare it on
                anchor_index: the anchor index to compare it on
                orientation: The orientation of the bar chart
            Returns:
                None, returns a horizontal barplot showing cosine similarity scores.

            Example:
                >>> cluster_field = 'question_vector_'
                >>> cluster_label = 'category'
                >>> alias = '1st_cluster'
                >>> dim_reduction_field = '_dr_.default.2.question_vectors_' # the '.' in names implies nested dictionaries in Vi
                >>> vi_client.plot_1d_cosine_similarity(collection=collection_name,
                        cluster_field=cluster_field,
                        cluster_label=cluster_label,
                        point_label='question_title',
                        dim_reduction_field=dim_reduction_field, 
                        include_centroids=True, 
                        alias=alias)
            
            """
            
            try:
                scores = self.get_cosine_similarity_scores(
                    documents, anchor_document, vector_field=vector_field
                )
            except NameError:
                raise APIError(f"Anchor document with {anchor_index} not found.")

            for i, doc in enumerate(documents):
                doc["cos_score"] = scores[i]

            mean_dict = MeanDict()
            for i, doc in enumerate(documents):
                mean_dict[doc[label]] = doc["cos_score"]
            
            x, y = mean_dict.get_x_y()
            if orientation == 'h':
                return go.Bar(
                    x=y,
                    y=x,
                    orientation=orientation,
                    name=vector_field
                )
            else:
                return go.Bar(
                    x=x,
                    y=y,
                    text=x,
                    orientation=orientation,
                    name=vector_field
                )
        
    def plot_2d_cosine_similarity(
        self,
        documents: List[Dict[str, Any]],
        anchor_documents: List[Dict[str, Any]],
        vector_fields: Union[str, List[str]],
        label: str,
        mode: str='markers+text'
    ):
        """
        Plot cosine similarity with 2 anchor documents to compare with against other documents.

        Args:
            documents: list of documents (dictionaries) to feed in
            vector_field: vector field to calculate cosine similarity on 
            label: the x label for the bar plot 
            anchor_document: the document to compare it on
            anchor_index: the anchor index to compare it on
        Returns:
            A scatterplot showing cosine similarity scores with each axes being a specific document.
        
        Example:
            >>> cluster_field = 'question_vector_'
            >>> cluster_label = 'category'
            >>> alias = '1st_cluster'
            >>> dim_reduction_field = '_dr_.default.2.question_vectors_' # the '.' in names implies nested dictionaries in Vi
            >>> vi_client.plot_2d_cosine_similarity(documents=documents[2:],
                    anchor_documents=documents[:2],
                    vector_fields=vector_field,
                    label=label
                    )

        """
        assert (
            len(anchor_documents) == 2
        ), "You need 2 anchor documents for a 2d cosine similarity plot."
        fig = go.Figure()
        if isinstance(vector_fields, str):
            vector_fields = [vector_fields]
        for vector_field in vector_fields:
            
            scores_x = self.get_cosine_similarity_scores(
                documents, anchor_documents[0], vector_field
            )
            scores_y = self.get_cosine_similarity_scores(
                documents, anchor_documents[1], vector_field
            )

            for i, doc in enumerate(documents):
                doc["cos_score_x"] = scores_x[i]
                doc["cos_score_y"] = scores_y[i]

            x = [round(x["cos_score_x"], 3) for x in documents]
            y = [round(x["cos_score_y"], 3) for x in documents]
            labels = [x[label] for x in documents]
            fig.add_trace(go.Scatter(x=x, y=y, text=labels, mode=mode, name=vector_field))
        fig.update_xaxes(title_text=f"Comparing with {anchor_documents[0][label]}")
        fig.update_yaxes(title_text=f"Comparing with {anchor_documents[1][label]}")
        fig.update_layout(
            title_text=f"2D Cosine Similarity Comparison With {anchor_documents[0][label]} and {anchor_documents[1][label]}"
        )
        return fig

    def _plot_radar(self, scores: list, spokes: list, name: str, fill=None):
        """
            Args:
                Scores: The list of scores for cosine similarity
                spokes: The outside labels 
                name: The name of the plot
        """
        return go.Scatterpolar(
            # Get the cosine similarity scores here 
            r=scores + [scores[0]],
            theta=spokes + [spokes[0]],
            fill=fill,
            name=name,
        )

    def plot_radar_across_documents(self, docs: List[Dict], anchor_documents: List[Dict], vector_field: str, 
    label_field: str, range: List=[0, 1], fill: str=None, scoring_metric ='cosine'):
        """
            Radar plot for 1D cosine similarity across documents.
            Args:
                docs: A list of documents 
                anchor_document: The document to compare against
                vector_field: The vector vector field
                label_field: The field of the documents to get labels.
        """
        categories = self.get_field_across_documents(label_field, docs)
        fig = go.Figure()
        for anchor_document in anchor_documents:
            if scoring_metric == 'cosine':
                scores = self.get_cosine_similarity_scores(docs, anchor_document, vector_field=vector_field)
            else:
                scores = scoring_metric(docs, anchor_document, vector_field=vector_field)
            fig.add_trace(self._plot_radar(scores=scores, spokes=categories, name=vector_field, fill=fill))
        
        fig.update_layout(
            polar=dict(
            radialaxis=dict(
                visible=True,
                range=range
            )),
            showlegend=True
        )
        return fig

    def plot_radar_across_vector_fields(self, docs: List[Dict], anchor_document: Dict, 
    vector_fields: List[str], label_field: str, range=[0, 1], fill=None, scoring_metric='cosine'):
        """
            Radar plot for 1D cosine similarity across different vector spaces.
            Args:
                docs: A list of documents 
                anchor_document: The document to compare against
                vector_fields: the different vector fields
                label_field: The field of the documents to get labels.
        """
        categories = self.get_field_across_documents(label_field, docs)
        fig = go.Figure()
        for vector in vector_fields:
            if scoring_metric == 'cosine':
                scores = self.get_cosine_similarity_scores(docs, anchor_document, vector_field=vector)
            else:
                scores = scoring_metric(docs, anchor_document, vector_field=vector)
            
            fig.add_trace(self._plot_radar(scores=scores, spokes=categories, name=vector, fill=fill))

        fig.update_layout(
            polar=dict(
            radialaxis=dict(
                visible=True,
                range=range
            )),
            showlegend=True 
        )
        return fig
